/* MIX simulator, copyright 1994 by Darius Bacon */ 
#include "mix.h"

#include "asm.h"
#include "driver.h"
#include "parse.h"
#include "symbol.h"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char label[33];

static Byte the_C, default_F;
static Flag F_must_be_default;

/* --- Handlers for each type of directive --- */

static void do_opcode(void)
{
    assemble(set_byte(the_C, 5, parse_operand(F_must_be_default, default_F)));
}

static void do_con(void)
{
    assemble(parse_W());
}

static void do_alf(void)
{
    error("Use CON instead of ALF");
}

static void do_end(void)
{
    set_entry_point(cell2address(parse_W()));
}

static void do_equ(void)
{
    define_symbol(string2symbol(label), parse_W());
    print_cell(symbol_value(string2symbol(label)));
    printf("\n");
}

static void do_orig(void)
{
    here = cell2address(parse_W());
}

/* --- The opcode/directive table --- */

typedef const struct OpDef {
    char name[5];       /* none of the names are > 4 chars long */
    void (*handler)(void);
    Byte C, F;
    Flag is_extended;
} OpDef;

#define extended true
#define def_opcode(name, C, F, is_ext)  { name, do_opcode, C, F, is_ext }
#define def_directive(name, handler)    { name, handler, 0, 0, 0 }

static OpDef op_table[] = {

#include "ops.inc"

};

/* --- The main driver --- */

static int delta_C;

static int compare_mnemonics(const char *datum, const char *pattern)
{
    unsigned my_delta_C = 0;
    for (; *pattern; ++pattern, ++datum) {
	char d = tolower(*datum);
	if (*pattern == '-') {
	    static const char legals[] = "a123456x";
	    const char *temp = strchr(legals, d);
	    if (temp)
		my_delta_C = temp - legals;
	    else
		break;
	} else if (*pattern == d)
	    ;
	else
	    break;
    }
    {
	int result = tolower(*datum) - *pattern;
	if (result == 0)
	    delta_C = my_delta_C;
	return result;
    }
}

typedef void (*Handler)(void);

/* I'd prefer something faster than a linear search here... */
static Handler lookup_mnemonic(const char *mnemonic)
{
    unsigned i;
    delta_C = 0;
    for (i = 0; i < sizeof op_table / sizeof op_table[0]; ++i) {
	if (compare_mnemonics(mnemonic, op_table[i].name) == 0) {
	    the_C = op_table[i].C + delta_C;
	    default_F = op_table[i].F;
	    F_must_be_default = op_table[i].is_extended;
	    return op_table[i].handler;
	}
    }
    return NULL;
}

static const char *skip_blanks(const char *s)
{
    return s + strspn(s, " \t");
}

typedef const char *const_char_ptr;

static void eat_identifier(char *buffer, const_char_ptr *s)
{
    char *scan = buffer;
    while (!isspace(**s) && **s != '\0') {
	if (!isalnum(**s))
	    error("Non-alphanumeric character in label or mnemonic");
	if (32 <= scan - buffer)
	    error("Label or mnemonic exceeds 32 characters");
	*scan++ = *(*s)++;
    }
    *scan = '\0';
}

void assemble_line(const char *line)
{
    char mnemonic[33];
    const char *scan = line;
    
    if (*skip_blanks(scan) == '*')      /* the comment character */
	return;

    eat_identifier(label, &scan);       /* eat label if any */
    scan = skip_blanks(scan);
    eat_identifier(mnemonic, &scan);    /* eat mnemonic */
    scan = skip_blanks(scan);
    setup_scanner(scan);                /* prepare to parse operand field */

    {
	Handler handler = lookup_mnemonic(mnemonic);
	if (handler != do_equ && label[0] != '\0')
	    define_symbol(string2symbol(label), address2cell(here));
	if (handler) {
	    handler();
	    done_parsing();  /* this goes here since *all* handlers parse an operand field */
	} else if (mnemonic[0] != '\0')
	    error("Unknown instruction: %s", mnemonic);
	else
	    ;
    }
}
